import re
from pygments.lexer import RegexLexer
from pygments.token import Text, Keyword


class GaqlLexer(RegexLexer):
    name = 'GAQL'
    aliases = 'gaql'
    filenames = ['*.gaql']

    flags = re.IGNORECASE
    tokens = {
        'root': [
            (r'\s+', Text),
            (
                r'\b(SELECT|FROM|LIMIT|WHERE|ORDER BY|BETWEEN|'
                r'LIKE|NOT LIKE|CONTAINS ANY|CONTAINS ALL|'
                r'CONTAINS NONE|IS NULL|IS NOT NULL|DURING|'
                r'AND|ASC|DESC|LAST_14_DAYS|LAST_30_DAYS|'
                r'LAST_7_DAYS|LAST_BUSINESS_WEEK|LAST_MONTH|'
                r'LAST_WEEK_MON_SUN|LAST_WEEK_SUN_SAT|THIS_MONTH|'
                r'THIS_WEEK_MON_TODAY|THIS_WEEK_SUN_TODAY|TODAY|'
                r'YESTERDAY)\b',
                Keyword,
            ),
        ]
    }



import pathlib

FORMATS = ['json', 'jsonl', 'csv', 'proto']
DEFAULT_LIMIT = '100'
USER_HOME = pathlib.Path.home()

CONFIG_DIR = USER_HOME / '.config/gaql/'
CREDENTIAL_FILE = CONFIG_DIR / 'credentials.json'
HISTORY_FILE = CONFIG_DIR / 'history'



import click
import json
import os
import sys

from gaql import CONFIG_DIR, CREDENTIAL_FILE


def configure():
    login_customer_id = input("login_customer_id: ").replace("-", "")

    secrets = {
        "developer_token": input("developer_token: "),
        "client_id": input("client_id: "),
        "client_secret": input("client_secret: "),
        "login_customer_id": login_customer_id,
        "refresh_token": input("refresh_token: "),
    }

    os.makedirs(CONFIG_DIR, exist_ok=True)

    with CREDENTIAL_FILE.open("w") as f:
        print(f"Wrote credentials to {CREDENTIAL_FILE}")
        f.write(json.dumps(secrets, indent=4))


def read_credentials_from_file():
    def read_credential_file_json():
        with CREDENTIAL_FILE.open("r") as f:
            return json.load(f)

    try:
        return read_credential_file_json()
    except FileNotFoundError:
        print(f"Configuration file not found at : {CREDENTIAL_FILE}", file=sys.stderr)
        should_configure = click.confirm("Configure it now?")
        if should_configure:
            configure()
            return read_credential_file_json()
        ERROR_MSG = (
            "Cannot continue without credentials. Please add credentials to your environment, home directory,"
            " or fill out the credentials prompt."
        )
        print(ERROR_MSG, file=sys.stderr)
        sys.exit(1)


def load_credentials():
    """Loads credentials with the following ordering: from env, Google's yaml file, a custom .json file in .config/gaql"""
    from google.ads.googleads import config

    try:
        return config.load_from_env()
    except:
        try:
            return config.load_from_yaml_file()
        except:
            if not CREDENTIAL_FILE.exists():
                print(
                    f"Couldn't load credentials from the environment, or from ~/google-ads.yaml. Trying {CREDENTIAL_FILE}"
                )
            file_credentials = read_credentials_from_file()
            file_credentials.update({"use_proto_plus": True})
            return config.load_from_dict(file_credentials)


def get_root_client():
    return load_credentials()["login_customer_id"]


def setup_client():
    from google.ads.googleads.client import GoogleAdsClient

    return GoogleAdsClient.load_from_dict(load_credentials())


def setup_ads_service(client):
    return client.get_service("GoogleAdsService", version="v9")


def setup_fields_service(client):
    return client.get_service("GoogleAdsFieldService", version="v9")



import os
import sys

import click

from gaql import FORMATS


class State(object):
    def __init__(self):
        self.output = sys.stdout
        self.format = None

    def finalize(self, query):
        self.__validate_output(query)
        self.format = self.__coerce_format()

    def __validate_output(self, query):
        if not self.output or self.output.name == '<stdout>':
            return

        if len(query) == 0:
            click.echo('[FATAL]: if output is set, a query must be provided', sys.stderr)
            sys.exit(1)

    def __coerce_format(self):
        """
        If a file format was provided, use that format, otherwise use the filename's format
        if it's a supported format. Otherwise, default to json
        """
        if self.format:
            return self.format

        filename = self.output.name
        if '.' in filename:
            ext = os.path.splitext(filename)[1][1:]
            if ext in FORMATS:
                return ext
        return 'json'


pass_state = click.make_pass_decorator(State, ensure=True)


def output_option(f):
    def callback(ctx, param, value):
        state = ctx.ensure_object(State)
        state.output = value
        return value

    return click.option(
        '-o',
        '--output',
        type=click.File('w'),
        help='allows outputting to a file',
        default='-',
        expose_value=False,
        callback=callback,
    )(f)


def format_option(f):
    def callback(ctx, param, value):
        state = ctx.ensure_object(State)
        state.format = value
        return value

    return click.option(
        '-f',
        '--format',
        type=click.Choice(['json', 'jsonl', 'csv', 'proto'], case_sensitive=False),
        help='the format to output rows in (defaults to json)',
        expose_value=False,
        callback=callback,
    )(f)


def common_options(f):
    f = output_option(f)
    f = format_option(f)
    return f



"""Script for maintaining autocomplete.py"""
from gaql.lib.click_decorators.state import State
from gaql.lib.google_clients.config import setup_client
from gaql.lib.google_clients.queries import google_fields_query
from gaql.lib.output import to_dict
import json
import pathlib


def attributes_query(resource):
    return f"SELECT name WHERE category='ATTRIBUTE' AND name LIKE '{resource}.%'"


def resources_query():
    return f"SELECT name, selectable_with WHERE category = 'RESOURCE'"


def autocomplete_fields():
    """Dumps entity fields from Google, building a single large JSON which provides the seed data for autocompletion"""
    state = State()
    state.format = 'json'

    client = setup_client()
    query_method = google_fields_query(client)

    resources_path = pathlib.Path('gaql/lib/google_clients/completion/entities.json')
    resource_rows = query_method(resources_query())
    resources = {resource.name: to_dict(resource._pb) for resource in resource_rows}
    if resources_path.exists():
        with resources_path.open('r') as f:
            existing_resources = json.load(f)
            for k in existing_resources:
                resources[k] = existing_resources[k]

    for resource_name, resource_dict in resources.items():
        if "fields" in resource_dict:
            continue
        print(f"Dumping {resource_name} to json")

        resource_dict["fields"] = []
        attributes = query_method(attributes_query(resource_name))
        for attribute in attributes:
            resource_dict["fields"].append(attribute.name)

    with resources_path.open('w') as f:
        json.dump(resources, fp=f)


def flatten_completion():
    """flattens all entities into a dictionary { name: fields }, with one aggregating entry { all: all_fields }"""
    resources_path = pathlib.Path('gaql/lib/google_clients/completion/entities.json')
    output_path = pathlib.Path('gaql/lib/google_clients/completion/autocompletions.py')

    with resources_path.open('r') as f:
        autocompletion = json.load(f)
    flattened_autocompletion = {}

    for k in autocompletion:
        raw_fields = autocompletion[k]["fields"] + autocompletion[k].get("selectable_with", [])
        flattened_fields = []

        for field in raw_fields:
            if "." in field:
                flattened_fields.append(field)
            else:
                flattened_fields += autocompletion[field]["fields"]
        flattened_autocompletion[k] = flattened_fields

    all_fields = set()
    for k in flattened_autocompletion:
        for field in flattened_autocompletion[k]:
            all_fields.add(field)

    flattened_autocompletion["all"] = list(sorted(all_fields))

    with output_path.open('w') as f:
        f.write("COMPLETIONS = ")
        json.dump(flattened_autocompletion, fp=f)


if __name__ == "__main__":
    autocomplete_fields()
    flatten_completion()



from gaql import DEFAULT_LIMIT
from gaql.lib.google_clients.config import setup_ads_service, setup_fields_service
from google.ads.googleads.client import GoogleAdsClient
from google.ads.googleads.errors import GoogleAdsException


def google_ads_query(client, account_id):
    service = setup_ads_service(client)

    def do_query(query):
        search_request = client.get_type("SearchGoogleAdsStreamRequest")
        search_request.customer_id = account_id
        search_request.query = query
        response = service.search_stream(search_request)
        for batch in response:
            for row in batch.results:
                yield row._pb

    return do_query


def google_fields_query(client):
    service = setup_fields_service(client)

    def do_query(query):
        return service.search_google_ads_fields(query=query)

    return do_query



from functools import reduce


class Trie(object):
    def __init__(self):
        self.children = {}
        self.terminator = False

    def add(self, char):
        self.children[char] = Trie()

    def insert(self, word):
        node = self
        for char in word:
            if char not in node.children:
                node.add(char)
            node = node.children[char]
        node.terminator = True

    def contains(self, word):
        node = self
        for char in word:
            if char not in node.children:
                return False
            node = node.children[char]
        return node.terminator

    def all_suffixes(self, prefix):
        results = []
        if self.terminator:
            results.append(prefix)
        if not self.children:
            return results
        return (
            reduce(
                lambda a, b: a + b,
                [node.all_suffixes(prefix + char) for (char, node) in self.children.items()],
            )
            + results
        )

    def autocomplete(self, prefix):
        node = self
        for char in prefix:
            if char not in node.children:
                return []
            node = node.children[char]
        return list(node.all_suffixes(prefix))
